from typing import Any, Iterator, List

def generate_combo(words: List[Any], min_length: int=..., max_length: int=..., join_by: str=...) -> Iterator[tuple]: ...
def hex_chars() -> list: ...
def generate_random_string(length: int=8) -> str: ...
