from ..core import ChepyCore
from typing import Any, TypeVar

NetworkingT = TypeVar('NetworkingT', bound='Networking')

class Networking(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def defang_url(self: NetworkingT) -> NetworkingT: ...
    def refang_url(self: NetworkingT) -> NetworkingT: ...
    def defang_ip(self: NetworkingT) -> NetworkingT: ...
    def refang_ip(self: NetworkingT) -> NetworkingT: ...
    def parse_uri(self: NetworkingT) -> NetworkingT: ...
    def parse_ip_range(self: NetworkingT) -> NetworkingT: ...
    def parse_ipv6(self: NetworkingT) -> NetworkingT: ...
    def int_to_ip(self: NetworkingT) -> NetworkingT: ...
    def ip_to_int(self: NetworkingT) -> NetworkingT: ...
    def get_ssl_cert(self: NetworkingT, port: int=...) -> NetworkingT: ...
