from ..core import ChepyCore
from typing import Any, TypeVar, Literal

pydash: Any
CodeTidyT = TypeVar('CodeTidyT', bound='CodeTidy')

class CodeTidy(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def minify_json(self: CodeTidyT) -> CodeTidyT: ...
    def beautify_json(self: CodeTidyT, indent: int=...) -> CodeTidyT: ...
    def to_upper_case(self: CodeTidyT, by: Literal['all', 'word', 'sentence']=...) -> CodeTidyT: ...
    def to_lower_case(self: CodeTidyT) -> CodeTidyT: ...
    def to_snake_case(self: CodeTidyT) -> CodeTidyT: ...
    def to_camel_case(self: CodeTidyT, ignore_space: bool=...) -> CodeTidyT: ...
    def to_kebab_case(self: CodeTidyT) -> CodeTidyT: ...
    def swap_case(self: CodeTidyT) -> CodeTidyT: ...
    def to_leetspeak(self: CodeTidyT, special_chars: bool=...) -> CodeTidyT: ...
    def random_case(self) -> CodeTidyT: ...
