from ..core import ChepyCore
from typing import Any, TypeVar, Union

SearchT = TypeVar('SearchT', bound='Search')

class Search(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def search(self: SearchT, pattern: Union[str, bytes]) -> SearchT: ...
    def search_list(self: SearchT, pattern: Union[str, bytes]) -> SearchT: ...
    def search_ctf_flags(self: SearchT, prefix: str, postfix: str=...) -> SearchT: ...
    def search_slack_tokens(self: SearchT) -> SearchT: ...
    def search_slack_webhook(self: SearchT) -> SearchT: ...
    def search_private_key(self: SearchT) -> SearchT: ...
    def search_twilio_key(self: SearchT) -> SearchT: ...
    def search_aws_key(self: SearchT) -> SearchT: ...
