from typing import Any

class ChepyConfig:
    chepy_dir: Any = ...
    chepy_conf: Any = ...
    config: Any = ...
    enable_plugins: Any = ...
    plugin_path: Any = ...
    history_path: Any = ...
    prompt_char: Any = ...
    prompt_colors: Any = ...
    show_rprompt: Any = ...
    prompt_rprompt: Any = ...
    prompt_bottom_toolbar: Any = ...
    prompt_toolbar_version: Any = ...
    prompt_toolbar_states: Any = ...
    prompt_toolbar_buffers: Any = ...
    prompt_toolbar_type: Any = ...
    prompt_toolbar_plugins: Any = ...
    prompt_toolbar_errors: Any = ...
    prompt_cli_method: Any = ...
    prompt_plugin_method: Any = ...
    cli_info_color: Any = ...
    def __init__(self) -> None: ...
    def load_plugins(self): ...
