from typing import Dict, Iterator, Union, Any

def factordb(n: int) -> dict: ...
def construct_private_key(n: int, e: int, d: int, format: str=..., passphrase: str=...) -> str: ...
def xor_bruteforce_multi(data: str, min: int=..., max: int=..., errors: str=...) -> Iterator[Dict[str, str]]: ...
def xor_repeating_key(data1:bytes, data2: bytes, min: int =..., max: int = ...) -> Union[bytes, None]: ...
def generate_rsa_keypair(bits: int=..., passphrase: str=...) -> Dict[str, Dict[str, Any]] : ...