from typing import Any, Dict, List

class Encoding:
    py_encodings: Any = ...
    py_text_encodings: Any = ...
    asciichars: Any = ...
    brailles: Any = ...
    wingdings: dict = ...
    NATO_CONSTANTS_DICT: Dict[str, str] = ...
    LEETCODE: List[List[str]] = ...
    BASE91_ALPHABET: List[str] = ...

class EncryptionConsts:
    MORSE_CODE_DICT: Any = ...

class PcapUSB:
    qwerty_map: Any = ...
    qwerty_modifier: Any = ...
    dvorak: Any = ...
    dvorak_modifier: Any = ...
