from prompt_toolkit.completion import Completer
from typing import Any

module: Any
options: Any
config: Any

class CliCompleter(Completer):
    def get_completions(self, document: Any, complete_event: Any) -> None: ...

def functions_cli(): ...
def get_doc(method: str) -> Any: ...
def cli_edit_state(fire: object, args: list) -> Any: ...
def cli_highlight(fire: object, highlight: str) -> Any: ...
def get_cli_options(): ...
def print_in_colors(out: Any) -> None: ...
def cli_state_type(fire: object) -> Any: ...
def cli_get_state(fire: object, index: int) -> Any: ...
def cli_show_length(fire: object) -> Any: ...
def cli_show_dict_keys(fire: object, pretty: bool=...) -> Any: ...
def cli_show_states(fire: object, pretty: bool=...) -> Any: ...
def cli_show_buffers(fire: object, pretty: bool=...) -> Any: ...
def cli_get_attr(fire: object, attr: str) -> Any: ...
def cli_pretty_print(fire: object) -> Any: ...
def cli_plugin_path(config: Any) -> None: ...
def cli_show_errors(errors: Any) -> None: ...
def cli_go_back() -> None: ...
def cli_delete_history() -> None: ...
def cli_exit(fire: object) -> Any: ...
