def yellow_background(s: str) -> str: ...
def red(s: str) -> str: ...
def blue(s: str) -> str: ...
def cyan(s: str) -> str: ...
def green(s: str) -> str: ...
def yellow(s: str) -> str: ...
def magenta(s: str) -> str: ...
