from ..core import ChepyCore
from typing import Any, TypeVar

DateTimeT = TypeVar("DateTimeT", bound="DateTime")

class DateTime(ChepyCore):
    def __init__(self, *data: Any) -> None: ...
    state: Any = ...
    def from_unix_timestamp(self: DateTimeT, format: str = ..., utc: bool = ...) -> DateTimeT: ...
    def to_unix_timestamp(self: DateTimeT) -> DateTimeT: ...
